@@ -15,12 +15,12 @@
 """Cursor class to iterate over Mongo query results."""
 
 import copy
-import datetime
+import threading
 import warnings
 
 from collections import deque
 
-from bson import RE_TYPE
+from bson import RE_TYPE, _convert_raw_document_lists_to_streams
 from bson.code import Code
 from bson.py3compat import (iteritems,
                             integer_types,
@@ -29,28 +29,54 @@ from bson.son import SON
 from pymongo import helpers
 from pymongo.common import validate_boolean, validate_is_mapping
 from pymongo.collation import validate_collation_or_none
-from pymongo.errors import (AutoReconnect,
-                            ConnectionFailure,
+from pymongo.errors import (ConnectionFailure,
                             InvalidOperation,
-                            NotMasterError,
                             OperationFailure)
-from pymongo.message import (_convert_exception,
-                             _CursorAddress,
+from pymongo.message import (_CursorAddress,
                              _GetMore,
                              _RawBatchGetMore,
                              _Query,
                              _RawBatchQuery)
+from pymongo.response import PinnedResponse
 
+# These errors mean that the server has already killed the cursor so there is
+# no need to send killCursors.
+_CURSOR_CLOSED_ERRORS = frozenset([
+    43,   # CursorNotFound
+    50,   # MaxTimeMSExpired
+    175,  # QueryPlanKilled
+    237,  # CursorKilled
+
+    # On a tailable cursor, the following errors mean the capped collection
+    # rolled over.
+    # MongoDB 2.6:
+    # {'$err': 'Runner killed during getMore', 'code': 28617, 'ok': 0}
+    28617,
+    # MongoDB 3.0:
+    # {'$err': 'getMore executor error: UnknownError no details available',
+    #  'code': 17406, 'ok': 0}
+    17406,
+    # MongoDB 3.2 + 3.4:
+    # {'ok': 0.0, 'errmsg': 'GetMore command executor error:
+    #  CappedPositionLost: CollectionScan died due to failure to restore
+    #  tailable cursor position. Last seen record id: RecordId(3)',
+    #  'code': 96}
+    96,
+    # MongoDB 3.6+:
+    # {'ok': 0.0, 'errmsg': 'errmsg: "CollectionScan died due to failure to
+    #  restore tailable cursor position. Last seen record id: RecordId(3)"',
+    #  'code': 136, 'codeName': 'CappedPositionLost'}
+    136,
+])
 
 _QUERY_OPTIONS = {
     "tailable_cursor": 2,
-    "slave_okay": 4,
+    "secondary_okay": 4,
     "oplog_replay": 8,
     "no_timeout": 16,
     "await_data": 32,
     "exhaust": 64,
     "partial": 128}
-_CURSOR_DOC_FIELDS = {'cursor': {'firstBatch': 1, 'nextBatch': 1}}
 
 
 class CursorType(object):
@@ -82,26 +108,25 @@ class CursorType(object):
     """
 
 
-# This has to be an old style class due to
-# http://bugs.jython.org/issue1057
-class _SocketManager:
+class _SocketManager(object):
     """Used with exhaust cursors to ensure the socket is returned.
     """
-    def __init__(self, sock, pool):
+    def __init__(self, sock, more_to_come):
         self.sock = sock
-        self.pool = pool
-        self.__closed = False
+        self.more_to_come = more_to_come
+        self.closed = False
+        self.lock = threading.Lock()
 
-    def __del__(self):
-        self.close()
+    def update_exhaust(self, more_to_come):
+        self.more_to_come = more_to_come
 
     def close(self):
         """Return this instance's socket to the connection pool.
         """
-        if not self.__closed:
-            self.__closed = True
-            self.pool.return_socket(self.sock)
-            self.sock, self.pool = None, None
+        if not self.closed:
+            self.closed = True
+            self.sock.unpin()
+            self.sock = None
 
 
 class Cursor(object):
@@ -116,20 +141,22 @@ class Cursor(object):
                  sort=None, allow_partial_results=False, oplog_replay=False,
                  modifiers=None, batch_size=0, manipulate=True,
                  collation=None, hint=None, max_scan=None, max_time_ms=None,
-                 max=None, min=None, return_key=False, show_record_id=False,
-                 snapshot=False, comment=None, session=None):
+                 max=None, min=None, return_key=None, show_record_id=None,
+                 snapshot=None, comment=None, session=None,
+                 allow_disk_use=None):
         """Create a new cursor.
 
         Should not be called directly by application developers - see
         :meth:`~pymongo.collection.Collection.find` instead.
 
-        .. mongodoc:: cursors
+        .. seealso:: The MongoDB documentation on `cursors <https://dochub.mongodb.org/core/cursors>`_.
         """
         # Initialize all attributes used in __del__ before possibly raising
         # an error to avoid attribute errors during garbage collection.
+        self.__collection = collection
         self.__id = None
         self.__exhaust = False
-        self.__exhaust_mgr = None
+        self.__sock_mgr = None
         self.__killed = False
 
         if session:
@@ -149,6 +176,14 @@ class Cursor(object):
         if not isinstance(limit, int):
             raise TypeError("limit must be an instance of int")
         validate_boolean("no_cursor_timeout", no_cursor_timeout)
+        if no_cursor_timeout and not self.__explicit_session:
+            warnings.warn("use an explicit session with no_cursor_timeout=True "
+                          "otherwise the cursor may still timeout after "
+                          "30 minutes, for more info see "
+                          "https://docs.mongodb.com/v4.4/reference/method/"
+                          "cursor.noCursorTimeout/"
+                          "#session-idle-timeout-overrides-nocursortimeout",
+                          UserWarning, stacklevel=2)
         if cursor_type not in (CursorType.NON_TAILABLE, CursorType.TAILABLE,
                                CursorType.TAILABLE_AWAIT, CursorType.EXHAUST):
             raise ValueError("not a valid value for cursor_type")
@@ -162,13 +197,15 @@ class Cursor(object):
             raise TypeError("batch_size must be an integer")
         if batch_size < 0:
             raise ValueError("batch_size must be >= 0")
+        # Only set if allow_disk_use is provided by the user, else None.
+        if allow_disk_use is not None:
+            allow_disk_use = validate_boolean("allow_disk_use", allow_disk_use)
 
         if projection is not None:
             if not projection:
                 projection = {"_id": 1}
             projection = helpers._fields_list_to_dict(projection, "projection")
 
-        self.__collection = collection
         self.__spec = spec
         self.__projection = projection
         self.__skip = skip
@@ -187,6 +224,7 @@ class Cursor(object):
         self.__collation = validate_collation_or_none(collation)
         self.__return_key = return_key
         self.__show_record_id = show_record_id
+        self.__allow_disk_use = allow_disk_use
         self.__snapshot = snapshot
         self.__set_hint(hint)
 
@@ -224,6 +262,10 @@ class Cursor(object):
         if oplog_replay:
             self.__query_flags |= _QUERY_OPTIONS["oplog_replay"]
 
+        # The namespace to use for find/getMore commands.
+        self.__dbname = collection.database.name
+        self.__collname = collection.name
+
     @property
     def collection(self):
         """The :class:`~pymongo.collection.Collection` that this
@@ -249,6 +291,7 @@ class Cursor(object):
         be sent to the server, even if the resultant data has already been
         retrieved by this cursor.
         """
+        self.close()
         self.__data = deque()
         self.__id = None
         self.__address = None
@@ -279,7 +322,9 @@ class Cursor(object):
                            "max_time_ms", "max_await_time_ms", "comment",
                            "max", "min", "ordering", "explain", "hint",
                            "batch_size", "max_scan", "manipulate",
-                           "query_flags", "modifiers", "collation")
+                           "query_flags", "modifiers", "collation", "empty",
+                           "show_record_id", "return_key", "allow_disk_use",
+                           "snapshot", "exhaust")
         data = dict((k, v) for k, v in iteritems(self.__dict__)
                     if k.startswith('_Cursor__') and k[9:] in values_to_clone)
         if deepcopy:
@@ -295,29 +340,31 @@ class Cursor(object):
     def __die(self, synchronous=False):
         """Closes this cursor.
         """
-        already_killed = self.__killed
+        try:
+            already_killed = self.__killed
+        except AttributeError:
+            # __init__ did not run to completion (or at all).
+            return
+
         self.__killed = True
         if self.__id and not already_killed:
-            if self.__exhaust and self.__exhaust_mgr:
-                # If this is an exhaust cursor and we haven't completely
-                # exhausted the result set we *must* close the socket
-                # to stop the server from sending more data.
-                self.__exhaust_mgr.sock.close()
-            else:
-                address = _CursorAddress(
-                    self.__address, self.__collection.full_name)
-                if synchronous:
-                    self.__collection.database.client._close_cursor_now(
-                        self.__id, address, session=self.__session)
-                else:
-                    # The cursor will be closed later in a different session.
-                    self.__collection.database.client._close_cursor(
-                        self.__id, address)
-        if self.__exhaust and self.__exhaust_mgr:
-            self.__exhaust_mgr.close()
-        if self.__session and not self.__explicit_session:
-            self.__session._end_session(lock=synchronous)
+            cursor_id = self.__id
+            address = _CursorAddress(
+                self.__address, "%s.%s" % (self.__dbname, self.__collname))
+        else:
+            # Skip killCursors.
+            cursor_id = 0
+            address = None
+        self.__collection.database.client._cleanup_cursor(
+            synchronous,
+            cursor_id,
+            address,
+            self.__sock_mgr,
+            self.__session,
+            self.__explicit_session)
+        if not self.__explicit_session:
             self.__session = None
+        self.__sock_mgr = None
 
     def close(self):
         """Explicitly close / kill this cursor.
@@ -344,19 +391,19 @@ class Cursor(object):
             operators["$max"] = self.__max
         if self.__min:
             operators["$min"] = self.__min
-        if self.__return_key:
+        if self.__return_key is not None:
             operators["$returnKey"] = self.__return_key
-        if self.__show_record_id:
+        if self.__show_record_id is not None:
             # This is upgraded to showRecordId for MongoDB 3.2+ "find" command.
             operators["$showDiskLoc"] = self.__show_record_id
-        if self.__snapshot:
+        if self.__snapshot is not None:
             operators["$snapshot"] = self.__snapshot
 
         if operators:
             # Make a shallow copy so we can cleanly rewind or clone.
             spec = self.__spec.copy()
 
-            # White-listed commands must be wrapped in $query.
+            # Allow-listed commands must be wrapped in $query.
             if "$query" not in spec:
                 # $query has to come first
                 spec = SON([("$query", spec)])
@@ -423,6 +470,28 @@ class Cursor(object):
         self.__query_flags &= ~mask
         return self
 
+    def allow_disk_use(self, allow_disk_use):
+        """Specifies whether MongoDB can use temporary disk files while
+        processing a blocking sort operation.
+
+        Raises :exc:`TypeError` if `allow_disk_use` is not a boolean.
+
+        .. note:: `allow_disk_use` requires server version **>= 4.4**
+
+        :Parameters:
+          - `allow_disk_use`: if True, MongoDB may use temporary
+            disk files to store data exceeding the system memory limit while
+            processing a blocking sort operation.
+
+        .. versionadded:: 3.11
+        """
+        if not isinstance(allow_disk_use, bool):
+            raise TypeError('allow_disk_use must be a bool')
+        self.__check_okay_to_chain()
+
+        self.__allow_disk_use = allow_disk_use
+        return self
+
     def limit(self, limit):
         """Limits the number of results to be returned by this cursor.
 
@@ -434,7 +503,7 @@ class Cursor(object):
         :Parameters:
           - `limit`: the number of results to return
 
-        .. mongodoc:: limit
+        .. seealso:: The MongoDB documentation on `limit <https://dochub.mongodb.org/core/limit>`_.
         """
         if not isinstance(limit, integer_types):
             raise TypeError("limit must be an integer")
@@ -547,6 +616,18 @@ class Cursor(object):
     def __getitem__(self, index):
         """Get a single document or a slice of documents from this cursor.
 
+        .. warning:: A :class:`~Cursor` is not a Python :class:`list`. Each
+          index access or slice requires that a new query be run using skip
+          and limit. Do not iterate the cursor using index accesses.
+          The following example is **extremely inefficient** and may return
+          surprising results::
+
+            cursor = db.collection.find()
+            # Warning: This runs a new query for each document.
+            # Don't do this!
+            for idx in range(10):
+                print(cursor[idx])
+
         Raises :class:`~pymongo.errors.InvalidOperation` if this
         cursor has already been used.
 
@@ -709,6 +790,9 @@ class Cursor(object):
             for doc in cursor:
                 print(doc)
 
+        For more advanced text search functionality, see MongoDB's
+        `Atlas Search <https://docs.atlas.mongodb.com/atlas-search/>`_.
+
         Raises :class:`~pymongo.errors.InvalidOperation` if this cursor has
         already been used. Only the last :meth:`sort` applied to this
         cursor has any effect.
@@ -823,7 +907,7 @@ class Cursor(object):
           :meth:`~pymongo.database.Database.command` to run the explain
           command directly.
 
-        .. mongodoc:: explain
+        .. seealso:: The MongoDB documentation on `explain <https://dochub.mongodb.org/core/explain>`_.
         """
         c = self.clone()
         c.__explain = True
@@ -887,7 +971,7 @@ class Cursor(object):
         return self
 
     def where(self, code):
-        """Adds a $where clause to this query.
+        """Adds a `$where`_ clause to this query.
 
         The `code` argument must be an instance of :class:`basestring`
         (:class:`str` in python 3) or :class:`~bson.code.Code`
@@ -895,7 +979,11 @@ class Cursor(object):
         evaluated for each document scanned. Only those documents
         for which the expression evaluates to *true* will be returned
         as results. The keyword *this* refers to the object currently
-        being scanned.
+        being scanned. For example::
+
+            # Find all documents where field "a" is less than "b" plus "c".
+            for doc in db.test.find().where('this.a < (this.b + this.c)'):
+                print(doc)
 
         Raises :class:`TypeError` if `code` is not an instance of
         :class:`basestring` (:class:`str` in python 3). Raises
@@ -903,8 +991,14 @@ class Cursor(object):
         :class:`Cursor` has already been used. Only the last call to
         :meth:`where` applied to a :class:`Cursor` has any effect.
 
+        .. note:: MongoDB 4.4 drops support for :class:`~bson.code.Code`
+          with scope variables. Consider using `$expr`_ instead.
+
         :Parameters:
           - `code`: JavaScript expression to use as a filter
+
+        .. _$expr: https://docs.mongodb.com/manual/reference/operator/query/expr/
+        .. _$where: https://docs.mongodb.com/manual/reference/operator/query/where/
         """
         self.__check_okay_to_chain()
         if not isinstance(code, Code):
@@ -941,134 +1035,54 @@ class Cursor(object):
         Can raise ConnectionFailure.
         """
         client = self.__collection.database.client
-        listeners = client._event_listeners
-        publish = listeners.enabled_for_commands
-        from_command = False
-        start = datetime.datetime.now()
-
-        def duration(): return datetime.datetime.now() - start
-
-        if operation:
-            try:
-                response = client._send_message_with_response(
-                    operation, exhaust=self.__exhaust, address=self.__address)
-                self.__address = response.address
-                if self.__exhaust:
-                    # 'response' is an ExhaustResponse.
-                    self.__exhaust_mgr = _SocketManager(response.socket_info,
-                                                        response.pool)
-
-                cmd_name = operation.name
-                reply = response.data
-                rqst_id = response.request_id
-                from_command = response.from_command
-            except AutoReconnect:
-                # Don't try to send kill cursors on another socket
-                # or to another server. It can cause a _pinValue
-                # assertion on some server releases if we get here
-                # due to a socket timeout.
-                self.__killed = True
-                self.__die()
-                raise
-        else:
-            # Exhaust cursor - no getMore message.
-            rqst_id = 0
-            cmd_name = 'getMore'
-            if publish:
-                # Fake a getMore command.
-                cmd = SON([('getMore', self.__id),
-                           ('collection', self.__collection.name)])
-                if self.__batch_size:
-                    cmd['batchSize'] = self.__batch_size
-                if self.__max_time_ms:
-                    cmd['maxTimeMS'] = self.__max_time_ms
-                listeners.publish_command_start(
-                    cmd, self.__collection.database.name, 0, self.__address)
-            try:
-                reply = self.__exhaust_mgr.sock.receive_message(None)
-            except Exception as exc:
-                if publish:
-                    listeners.publish_command_failure(
-                        duration(), _convert_exception(exc), cmd_name, rqst_id,
-                        self.__address)
-                if isinstance(exc, ConnectionFailure):
-                    self.__die()
-                raise
+        # OP_MSG is required to support exhaust cursors with encryption.
+        if client._encrypter and self.__exhaust:
+            raise InvalidOperation(
+                "exhaust cursors do not support auto encryption")
 
         try:
-            user_fields = None
-            legacy_response = True
-            if from_command:
-                user_fields = _CURSOR_DOC_FIELDS
-                legacy_response = False
-            docs = self._unpack_response(
-                reply, self.__id, self.__collection.codec_options,
-                legacy_response=legacy_response, user_fields=user_fields)
-            if from_command:
-                first = docs[0]
-                client._receive_cluster_time(first, self.__session)
-                helpers._check_command_response(first)
+            response = client._run_operation(
+                operation, self._unpack_response, address=self.__address)
         except OperationFailure as exc:
-            self.__killed = True
-
-            # Make sure exhaust socket is returned immediately, if necessary.
-            self.__die()
-
-            if publish:
-                listeners.publish_command_failure(
-                    duration(), exc.details, cmd_name, rqst_id, self.__address)
-
+            if exc.code in _CURSOR_CLOSED_ERRORS or self.__exhaust:
+                # Don't send killCursors because the cursor is already closed.
+                self.__killed = True
+            self.close()
             # If this is a tailable cursor the error is likely
             # due to capped collection roll over. Setting
             # self.__killed to True ensures Cursor.alive will be
             # False. No need to re-raise.
-            if self.__query_flags & _QUERY_OPTIONS["tailable_cursor"]:
+            if (exc.code in _CURSOR_CLOSED_ERRORS and
+                    self.__query_flags & _QUERY_OPTIONS["tailable_cursor"]):
                 return
             raise
-        except NotMasterError as exc:
-            # Don't send kill cursors to another server after a "not master"
-            # error. It's completely pointless.
+        except ConnectionFailure:
+            # Don't send killCursors because the cursor is already closed.
             self.__killed = True
-
-            # Make sure exhaust socket is returned immediately, if necessary.
-            self.__die()
-
-            if publish:
-                listeners.publish_command_failure(
-                    duration(), exc.details, cmd_name, rqst_id, self.__address)
-
-            client._reset_server_and_request_check(self.__address)
+            self.close()
             raise
-        except Exception as exc:
-            if publish:
-                listeners.publish_command_failure(
-                    duration(), _convert_exception(exc), cmd_name, rqst_id,
-                    self.__address)
+        except Exception:
+            self.close()
             raise
 
-        if publish:
-            # Must publish in find / getMore / explain command response format.
-            if from_command:
-                res = docs[0]
-            elif cmd_name == "explain":
-                res = docs[0] if docs else {}
-            else:
-                res = {"cursor": {"id": reply.cursor_id,
-                                  "ns": self.__collection.full_name},
-                       "ok": 1}
-                if cmd_name == "find":
-                    res["cursor"]["firstBatch"] = docs
-                else:
-                    res["cursor"]["nextBatch"] = docs
-            listeners.publish_command_success(
-                duration(), res, cmd_name, rqst_id, self.__address)
+        self.__address = response.address
+        if isinstance(response, PinnedResponse):
+            if not self.__sock_mgr:
+                self.__sock_mgr = _SocketManager(response.socket_info,
+                                                 response.more_to_come)
 
-        if from_command:
+        cmd_name = operation.name
+        docs = response.docs
+        if response.from_command:
             if cmd_name != "explain":
                 cursor = docs[0]['cursor']
                 self.__id = cursor['id']
                 if cmd_name == 'find':
                     documents = cursor['firstBatch']
+                    # Update the namespace used for future getMore commands.
+                    ns = cursor.get('ns')
+                    if ns:
+                        self.__dbname, self.__collname = ns.split('.', 1)
                 else:
                     documents = cursor['nextBatch']
                 self.__data = deque(documents)
@@ -1078,18 +1092,17 @@ class Cursor(object):
                 self.__data = deque(docs)
                 self.__retrieved += len(docs)
         else:
-            self.__id = reply.cursor_id
+            self.__id = response.data.cursor_id
             self.__data = deque(docs)
-            self.__retrieved += reply.number_returned
+            self.__retrieved += response.data.number_returned
 
         if self.__id == 0:
-            self.__killed = True
             # Don't wait for garbage collection to call __del__, return the
             # socket and the session to the pool now.
-            self.__die()
+            self.close()
 
         if self.__limit and self.__id and self.__limit <= self.__retrieved:
-            self.__die()
+            self.close()
 
     def _unpack_response(self, response, cursor_id, codec_options,
                          user_fields=None, legacy_response=False):
@@ -1136,7 +1149,9 @@ class Cursor(object):
                                   self.__read_concern,
                                   self.__collation,
                                   self.__session,
-                                  self.__collection.database.client)
+                                  self.__collection.database.client,
+                                  self.__allow_disk_use,
+                                  self.__exhaust)
             self.__send_message(q)
         elif self.__id:  # Get More
             if self.__limit:
@@ -1145,21 +1160,19 @@ class Cursor(object):
                     limit = min(limit, self.__batch_size)
             else:
                 limit = self.__batch_size
-
             # Exhaust cursors don't send getMore messages.
-            if self.__exhaust:
-                self.__send_message(None)
-            else:
-                g = self._getmore_class(self.__collection.database.name,
-                                        self.__collection.name,
-                                        limit,
-                                        self.__id,
-                                        self.__codec_options,
-                                        self._read_preference(),
-                                        self.__session,
-                                        self.__collection.database.client,
-                                        self.__max_await_time_ms)
-                self.__send_message(g)
+            g = self._getmore_class(self.__dbname,
+                                    self.__collname,
+                                    limit,
+                                    self.__id,
+                                    self.__codec_options,
+                                    self._read_preference(),
+                                    self.__session,
+                                    self.__collection.database.client,
+                                    self.__max_await_time_ms,
+                                    self.__sock_mgr,
+                                    self.__exhaust)
+            self.__send_message(g)
 
         return len(self.__data)
 
@@ -1300,7 +1313,7 @@ class RawBatchCursor(Cursor):
         see :meth:`~pymongo.collection.Collection.find_raw_batches`
         instead.
 
-        .. mongodoc:: cursors
+        .. seealso:: The MongoDB documentation on `cursors <https://dochub.mongodb.org/core/cursors>`_.
         """
         manipulate = kwargs.get('manipulate')
         kwargs['manipulate'] = False
@@ -1313,12 +1326,18 @@ class RawBatchCursor(Cursor):
 
     def _unpack_response(self, response, cursor_id, codec_options,
                          user_fields=None, legacy_response=False):
-        return response.raw_response(cursor_id)
+        raw_response = response.raw_response(
+            cursor_id, user_fields=user_fields)
+        if not legacy_response:
+            # OP_MSG returns firstBatch/nextBatch documents as a BSON array
+            # Re-assemble the array of documents into a document stream
+            _convert_raw_document_lists_to_streams(raw_response[0])
+        return raw_response
 
     def explain(self):
         """Returns an explain plan record for this cursor.
 
-        .. mongodoc:: explain
+        .. seealso:: The MongoDB documentation on `explain <https://dochub.mongodb.org/core/explain>`_.
         """
         clone = self._clone(deepcopy=True, base=Cursor(self.collection))
         return clone.explain()
