@@ -12,12 +12,15 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""Tools for creating `messages
+"""**DEPRECATED** Tools for creating `messages
 <http://www.mongodb.org/display/DOCS/Mongo+Wire+Protocol>`_ to be sent to
 MongoDB.
 
 .. note:: This module is for internal use and is generally not needed by
    application developers.
+
+.. versionchanged:: 3.12
+  This module is deprecated and will be removed in PyMongo 4.0.
 """
 
 import datetime
@@ -26,10 +29,15 @@ import struct
 
 import bson
 from bson import (CodecOptions,
-                  _bson_to_dict,
+                  decode,
+                  encode,
+                  _decode_selective,
                   _dict_to_bson,
                   _make_c_string)
 from bson.codec_options import DEFAULT_CODEC_OPTIONS
+from bson.int64 import Int64
+from bson.raw_bson import (_inflate_bson, DEFAULT_RAW_BSON_OPTIONS,
+                           RawBSONDocument)
 from bson.py3compat import b, StringIO
 from bson.son import SON
 
@@ -43,11 +51,13 @@ from pymongo.errors import (ConfigurationError,
                             DocumentTooLarge,
                             ExecutionTimeout,
                             InvalidOperation,
-                            NotMasterError,
+                            NotPrimaryError,
                             OperationFailure,
                             ProtocolError)
+from pymongo.hello_compat import HelloCompat
 from pymongo.read_concern import DEFAULT_READ_CONCERN
 from pymongo.read_preferences import ReadPreference
+from pymongo.write_concern import WriteConcern
 
 
 MAX_INT32 = 2147483647
@@ -92,21 +102,18 @@ def _randint():
 def _maybe_add_read_preference(spec, read_preference):
     """Add $readPreference to spec when appropriate."""
     mode = read_preference.mode
-    tag_sets = read_preference.tag_sets
-    max_staleness = read_preference.max_staleness
+    document = read_preference.document
     # Only add $readPreference if it's something other than primary to avoid
     # problems with mongos versions that don't support read preferences. Also,
     # for maximum backwards compatibility, don't add $readPreference for
     # secondaryPreferred unless tags or maxStalenessSeconds are in use (setting
-    # the slaveOkay bit has the same effect).
+    # the secondaryOkay bit has the same effect).
     if mode and (
-        mode != ReadPreference.SECONDARY_PREFERRED.mode
-        or tag_sets != [{}]
-        or max_staleness != -1):
-
+            mode != ReadPreference.SECONDARY_PREFERRED.mode or
+            len(document) > 1):
         if "$query" not in spec:
             spec = SON([("$query", spec)])
-        spec["$readPreference"] = read_preference.document
+        spec["$readPreference"] = document
     return spec
 
 
@@ -117,7 +124,7 @@ def _convert_exception(exception):
 
 
 def _convert_write_result(operation, command, result):
-    """Convert a legacy write result to write commmand format."""
+    """Convert a legacy write result to write command format."""
 
     # Based on _merge_legacy from bulk.py
     affected = result.get("n", 0)
@@ -179,7 +186,8 @@ _MODIFIERS = SON([
 
 
 def _gen_find_command(coll, spec, projection, skip, limit, batch_size, options,
-                      read_concern, collation=None, session=None):
+                      read_concern, collation=None, session=None,
+                      allow_disk_use=None):
     """Generate a find command document."""
     cmd = SON([('find', coll)])
     if '$query' in spec:
@@ -202,14 +210,17 @@ def _gen_find_command(coll, spec, projection, skip, limit, batch_size, options,
             cmd['singleBatch'] = True
     if batch_size:
         cmd['batchSize'] = batch_size
-    if read_concern.level and not (session and session._in_transaction):
+    if read_concern.level and not (session and session.in_transaction):
         cmd['readConcern'] = read_concern.document
     if collation:
         cmd['collation'] = collation
+    if allow_disk_use is not None:
+        cmd['allowDiskUse'] = allow_disk_use
     if options:
         cmd.update([(opt, True)
                     for opt, val in _OPTIONS.items()
                     if options & val])
+
     return cmd
 
 
@@ -230,11 +241,17 @@ class _Query(object):
     __slots__ = ('flags', 'db', 'coll', 'ntoskip', 'spec',
                  'fields', 'codec_options', 'read_preference', 'limit',
                  'batch_size', 'name', 'read_concern', 'collation',
-                 'session', 'client', '_as_command')
+                 'session', 'client', 'allow_disk_use', '_as_command',
+                 'exhaust')
+
+    # For compatibility with the _GetMore class.
+    sock_mgr = None
+    cursor_id = None
 
     def __init__(self, flags, db, coll, ntoskip, spec, fields,
                  codec_options, read_preference, limit,
-                 batch_size, read_concern, collation, session, client):
+                 batch_size, read_concern, collation, session, client,
+                 allow_disk_use, exhaust):
         self.flags = flags
         self.db = db
         self.coll = coll
@@ -249,14 +266,21 @@ class _Query(object):
         self.collation = collation
         self.session = session
         self.client = client
+        self.allow_disk_use = allow_disk_use
         self.name = 'find'
         self._as_command = None
+        self.exhaust = exhaust
 
-    def use_command(self, sock_info, exhaust):
+    def namespace(self):
+        return _UJOIN % (self.db, self.coll)
+
+    def use_command(self, sock_info):
         use_find_cmd = False
-        if sock_info.max_wire_version >= 4:
-            if not exhaust:
-                use_find_cmd = True
+        if sock_info.max_wire_version >= 4 and not self.exhaust:
+            use_find_cmd = True
+        elif sock_info.max_wire_version >= 8:
+            # OP_MSG supports exhaust on MongoDB 4.2+
+            use_find_cmd = True
         elif not self.read_concern.ok_for_legacy:
             raise ConfigurationError(
                 'read concern level of %s is not valid '
@@ -269,6 +293,11 @@ class _Query(object):
                 'Specifying a collation is unsupported with a max wire '
                 'version of %d.' % (sock_info.max_wire_version,))
 
+        if sock_info.max_wire_version < 4 and self.allow_disk_use is not None:
+            raise ConfigurationError(
+                'Specifying allowDiskUse is unsupported with a max wire '
+                'version of %d.' % (sock_info.max_wire_version,))
+
         sock_info.validate_session(self.client, self.session)
 
         return use_find_cmd
@@ -284,33 +313,36 @@ class _Query(object):
         cmd = _gen_find_command(
             self.coll, self.spec, self.fields, self.ntoskip,
             self.limit, self.batch_size, self.flags, self.read_concern,
-            self.collation, self.session)
+            self.collation, self.session, self.allow_disk_use)
         if explain:
             self.name = 'explain'
             cmd = SON([('explain', cmd)])
         session = self.session
+        sock_info.add_server_api(cmd)
         if session:
-            session._apply_to(cmd, False, self.read_preference)
+            session._apply_to(cmd, False, self.read_preference, sock_info)
             # Explain does not support readConcern.
-            if (not explain and session.options.causal_consistency
-                    and session.operation_time is not None
-                    and not session._in_transaction):
-                cmd.setdefault(
-                    'readConcern', {})[
-                    'afterClusterTime'] = session.operation_time
+            if not explain and not session.in_transaction:
+                session._update_read_concern(cmd, sock_info)
         sock_info.send_cluster_time(cmd, session, self.client)
+        # Support auto encryption
+        client = self.client
+        if (client._encrypter and
+                not client._encrypter._bypass_auto_encryption):
+            cmd = client._encrypter.encrypt(
+                self.db, cmd, False, self.codec_options)
         self._as_command = cmd, self.db
         return self._as_command
 
-    def get_message(self, set_slave_ok, sock_info, use_cmd=False):
-        """Get a query message, possibly setting the slaveOk bit."""
-        if set_slave_ok:
-            # Set the slaveOk bit.
+    def get_message(self, set_secondary_ok, sock_info, use_cmd=False):
+        """Get a query message, possibly setting the secondaryOk bit."""
+        if set_secondary_ok:
+            # Set the secondaryOk bit.
             flags = self.flags | 4
         else:
             flags = self.flags
 
-        ns = _UJOIN % (self.db, self.coll)
+        ns = self.namespace()
         spec = self.spec
 
         if use_cmd:
@@ -318,7 +350,7 @@ class _Query(object):
             if sock_info.op_msg_enabled:
                 request_id, msg, size, _ = _op_msg(
                     0, spec, self.db, self.read_preference,
-                    set_slave_ok, False, self.codec_options,
+                    set_secondary_ok, False, self.codec_options,
                     ctx=sock_info.compression_context)
                 return request_id, msg, size
             ns = _UJOIN % (self.db, "$cmd")
@@ -348,12 +380,13 @@ class _GetMore(object):
 
     __slots__ = ('db', 'coll', 'ntoreturn', 'cursor_id', 'max_await_time_ms',
                  'codec_options', 'read_preference', 'session', 'client',
-                 '_as_command')
+                 'sock_mgr', '_as_command', 'exhaust')
 
     name = 'getMore'
 
     def __init__(self, db, coll, ntoreturn, cursor_id, codec_options,
-                 read_preference, session, client, max_await_time_ms=None):
+                 read_preference, session, client, max_await_time_ms,
+                 sock_mgr, exhaust):
         self.db = db
         self.coll = coll
         self.ntoreturn = ntoreturn
@@ -363,11 +396,23 @@ class _GetMore(object):
         self.session = session
         self.client = client
         self.max_await_time_ms = max_await_time_ms
+        self.sock_mgr = sock_mgr
         self._as_command = None
+        self.exhaust = exhaust
+
+    def namespace(self):
+        return _UJOIN % (self.db, self.coll)
+
+    def use_command(self, sock_info):
+        use_cmd = False
+        if sock_info.max_wire_version >= 4 and not self.exhaust:
+            use_cmd = True
+        elif sock_info.max_wire_version >= 8:
+            # OP_MSG supports exhaust on MongoDB 4.2+
+            use_cmd = True
 
-    def use_command(self, sock_info, exhaust):
         sock_info.validate_session(self.client, self.session)
-        return sock_info.max_wire_version >= 4 and not exhaust
+        return use_cmd
 
     def as_command(self, sock_info):
         """Return a getMore command document for this query."""
@@ -380,22 +425,33 @@ class _GetMore(object):
                                     self.max_await_time_ms)
 
         if self.session:
-            self.session._apply_to(cmd, False, self.read_preference)
+            self.session._apply_to(cmd, False, self.read_preference, sock_info)
+        sock_info.add_server_api(cmd)
         sock_info.send_cluster_time(cmd, self.session, self.client)
+        # Support auto encryption
+        client = self.client
+        if (client._encrypter and
+                not client._encrypter._bypass_auto_encryption):
+            cmd = client._encrypter.encrypt(
+                self.db, cmd, False, self.codec_options)
         self._as_command = cmd, self.db
         return self._as_command
 
     def get_message(self, dummy0, sock_info, use_cmd=False):
         """Get a getmore message."""
 
-        ns = _UJOIN % (self.db, self.coll)
+        ns = self.namespace()
         ctx = sock_info.compression_context
 
         if use_cmd:
             spec = self.as_command(sock_info)[0]
             if sock_info.op_msg_enabled:
+                if self.sock_mgr:
+                    flags = _OpMsg.EXHAUST_ALLOWED
+                else:
+                    flags = 0
                 request_id, msg, size, _ = _op_msg(
-                    0, spec, self.db, ReadPreference.PRIMARY,
+                    flags, spec, self.db, None,
                     False, False, self.codec_options,
                     ctx=sock_info.compression_context)
                 return request_id, msg, size
@@ -405,29 +461,29 @@ class _GetMore(object):
         return get_more(ns, self.ntoreturn, self.cursor_id, ctx)
 
 
-# TODO: Use OP_MSG once the server is able to respond with document streams.
 class _RawBatchQuery(_Query):
-    def use_command(self, socket_info, exhaust):
+    def use_command(self, sock_info):
         # Compatibility checks.
-        super(_RawBatchQuery, self).use_command(socket_info, exhaust)
-
+        super(_RawBatchQuery, self).use_command(sock_info)
+        if sock_info.max_wire_version >= 8:
+            # MongoDB 4.2+ supports exhaust over OP_MSG
+            return True
+        elif sock_info.op_msg_enabled and not self.exhaust:
+            return True
         return False
 
-    def get_message(self, set_slave_ok, sock_info, use_cmd=False):
-        # Always pass False for use_cmd.
-        return super(_RawBatchQuery, self).get_message(
-            set_slave_ok, sock_info, False)
-
 
 class _RawBatchGetMore(_GetMore):
-    def use_command(self, socket_info, exhaust):
+    def use_command(self, sock_info):
+        # Compatibility checks.
+        super(_RawBatchGetMore, self).use_command(sock_info)
+        if sock_info.max_wire_version >= 8:
+            # MongoDB 4.2+ supports exhaust over OP_MSG
+            return True
+        elif sock_info.op_msg_enabled and not self.exhaust:
+            return True
         return False
 
-    def get_message(self, set_slave_ok, sock_info, use_cmd=False):
-        # Always pass False for use_cmd.
-        return super(_RawBatchGetMore, self).get_message(
-            set_slave_ok, sock_info, False)
-
 
 class _CursorAddress(tuple):
     """The server address (host, port) of a cursor, with namespace property."""
@@ -546,7 +602,11 @@ if _use_c:
 
 def insert(collection_name, docs, check_keys,
            safe, last_error_args, continue_on_error, opts, ctx=None):
-    """Get an **insert** message."""
+    """**DEPRECATED** Get an **insert** message.
+
+    .. versionchanged:: 3.12
+      This function is deprecated and will be removed in PyMongo 4.0.
+    """
     if ctx:
         return _insert_compressed(
             collection_name, docs, check_keys, continue_on_error, opts, ctx)
@@ -596,7 +656,11 @@ if _use_c:
 
 def update(collection_name, upsert, multi, spec,
            doc, safe, last_error_args, check_keys, opts, ctx=None):
-    """Get an **update** message."""
+    """**DEPRECATED** Get an **update** message.
+
+    .. versionchanged:: 3.12
+      This function is deprecated and will be removed in PyMongo 4.0.
+    """
     if ctx:
         return _update_compressed(
             collection_name, upsert, multi, spec, doc, check_keys, opts, ctx)
@@ -654,12 +718,13 @@ if _use_c:
     _op_msg_uncompressed = _cmessage._op_msg
 
 
-def _op_msg(flags, command, dbname, read_preference, slave_ok, check_keys,
+def _op_msg(flags, command, dbname, read_preference, secondary_ok, check_keys,
             opts, ctx=None):
     """Get a OP_MSG message."""
     command['$db'] = dbname
-    if "$readPreference" not in command:
-        if slave_ok and not read_preference.mode:
+    # getMore commands do not send $readPreference.
+    if read_preference is not None and "$readPreference" not in command:
+        if secondary_ok and not read_preference.mode:
             command["$readPreference"] = (
                 ReadPreference.PRIMARY_PREFERRED.document)
         else:
@@ -738,7 +803,11 @@ if _use_c:
 
 def query(options, collection_name, num_to_skip, num_to_return,
           query, field_selector, opts, check_keys=False, ctx=None):
-    """Get a **query** message."""
+    """**DEPRECATED** Get a **query** message.
+
+    .. versionchanged:: 3.12
+      This function is deprecated and will be removed in PyMongo 4.0.
+    """
     if ctx:
         return _query_compressed(options, collection_name, num_to_skip,
                                  num_to_return, query, field_selector,
@@ -775,7 +844,11 @@ if _use_c:
 
 
 def get_more(collection_name, num_to_return, cursor_id, ctx=None):
-    """Get a **getMore** message."""
+    """**DEPRECATED** Get a **getMore** message.
+
+    .. versionchanged:: 3.12
+      This function is deprecated and will be removed in PyMongo 4.0.
+    """
     if ctx:
         return _get_more_compressed(
             collection_name, num_to_return, cursor_id, ctx)
@@ -812,12 +885,15 @@ def _delete_uncompressed(
 
 def delete(
         collection_name, spec, safe, last_error_args, opts, flags=0, ctx=None):
-    """Get a **delete** message.
+    """**DEPRECATED** Get a **delete** message.
 
     `opts` is a CodecOptions. `flags` is a bit vector that may contain
     the SingleRemove flag or not:
 
     http://docs.mongodb.org/meta-driver/latest/legacy/mongodb-wire-protocol/#op-delete
+
+    .. versionchanged:: 3.12
+      This function is deprecated and will be removed in PyMongo 4.0.
     """
     if ctx:
         return _delete_compressed(collection_name, spec, opts, flags, ctx)
@@ -826,7 +902,10 @@ def delete(
 
 
 def kill_cursors(cursor_ids):
-    """Get a **killCursors** message.
+    """**DEPRECATED** Get a **killCursors** message.
+
+    .. versionchanged:: 3.12
+      This function is deprecated and will be removed in PyMongo 4.0.
     """
     num_cursors = len(cursor_ids)
     pack = struct.Struct("<ii" + ("q" * num_cursors)).pack
@@ -837,23 +916,55 @@ def kill_cursors(cursor_ids):
 class _BulkWriteContext(object):
     """A wrapper around SocketInfo for use with write splitting functions."""
 
-    __slots__ = ('db_name', 'command', 'sock_info', 'op_id',
+    __slots__ = ('db_name', 'sock_info', 'op_id',
                  'name', 'field', 'publish', 'start_time', 'listeners',
-                 'session', 'compress')
+                 'session', 'compress', 'op_type', 'codec', 'cmd_legacy')
 
-    def __init__(self, database_name, command, sock_info, operation_id,
-                 listeners, session):
+    def __init__(self, database_name, cmd_name, sock_info, operation_id,
+                 listeners, session, op_type, codec, cmd_legacy=None):
         self.db_name = database_name
-        self.command = command
         self.sock_info = sock_info
         self.op_id = operation_id
         self.listeners = listeners
         self.publish = listeners.enabled_for_commands
-        self.name = next(iter(command))
+        self.name = cmd_name
         self.field = _FIELD_MAP[self.name]
         self.start_time = datetime.datetime.now() if self.publish else None
         self.session = session
         self.compress = True if sock_info.compression_context else False
+        self.op_type = op_type
+        self.codec = codec
+        self.cmd_legacy = cmd_legacy
+
+    def _batch_command(self, cmd, docs):
+        namespace = self.db_name + '.$cmd'
+        request_id, msg, to_send = _do_bulk_write_command(
+            namespace, self.op_type, cmd, docs, self.check_keys,
+            self.codec, self)
+        if not to_send:
+            raise InvalidOperation("cannot do an empty bulk write")
+        return request_id, msg, to_send
+
+    def execute(self, cmd, docs, client):
+        request_id, msg, to_send = self._batch_command(cmd, docs)
+        result = self.write_command(cmd, request_id, msg, to_send)
+        client._process_response(result, self.session)
+        return result, to_send
+
+    def execute_unack(self, cmd, docs, client):
+        request_id, msg, to_send = self._batch_command(cmd, docs)
+        # Though this isn't strictly a "legacy" write, the helper
+        # handles publishing commands and sending our message
+        # without receiving a result. Send 0 for max_doc_size
+        # to disable size checking. Size checking is handled while
+        # the documents are encoded to BSON.
+        self.legacy_write(cmd, request_id, msg, 0, False, to_send)
+        return to_send
+
+    @property
+    def check_keys(self):
+        """Should we check keys for this operation type?"""
+        return False
 
     @property
     def max_bson_size(self):
@@ -863,6 +974,9 @@ class _BulkWriteContext(object):
     @property
     def max_message_size(self):
         """A proxy for SockInfo.max_message_size."""
+        if self.compress:
+            # Subtract 16 bytes for the message header.
+            return self.sock_info.max_message_size - 16
         return self.sock_info.max_message_size
 
     @property
@@ -870,20 +984,27 @@ class _BulkWriteContext(object):
         """A proxy for SockInfo.max_write_batch_size."""
         return self.sock_info.max_write_batch_size
 
+    @property
+    def max_split_size(self):
+        """The maximum size of a BSON command before batch splitting."""
+        return self.max_bson_size
+
     def legacy_bulk_insert(
             self, request_id, msg, max_doc_size, acknowledged, docs, compress):
         if compress:
             request_id, msg = _compress(
                 2002, msg, self.sock_info.compression_context)
         return self.legacy_write(
-            request_id, msg, max_doc_size, acknowledged, docs)
+            self.cmd_legacy.copy(), request_id, msg, max_doc_size,
+            acknowledged, docs)
 
-    def legacy_write(self, request_id, msg, max_doc_size, acknowledged, docs):
+    def legacy_write(self, cmd, request_id, msg, max_doc_size, acknowledged,
+                     docs):
         """A proxy for SocketInfo.legacy_write that handles event publishing.
         """
         if self.publish:
             duration = datetime.datetime.now() - self.start_time
-            cmd = self._start(request_id, docs)
+            cmd = self._start(cmd, request_id, docs)
             start = datetime.datetime.now()
         try:
             result = self.sock_info.legacy_write(
@@ -896,60 +1017,114 @@ class _BulkWriteContext(object):
                     # Comply with APM spec.
                     reply = {'ok': 1}
                 self._succeed(request_id, reply, duration)
-        except OperationFailure as exc:
+        except Exception as exc:
             if self.publish:
                 duration = (datetime.datetime.now() - start) + duration
-                self._fail(
-                    request_id,
-                    _convert_write_result(
-                        self.name, cmd, exc.details),
-                    duration)
+                if isinstance(exc, OperationFailure):
+                    failure = _convert_write_result(
+                        self.name, cmd, exc.details)
+                elif isinstance(exc, NotPrimaryError):
+                    failure = exc.details
+                else:
+                    failure = _convert_exception(exc)
+                self._fail(request_id, failure, duration)
             raise
         finally:
             self.start_time = datetime.datetime.now()
         return result
 
-    def write_command(self, request_id, msg, docs):
+    def write_command(self, cmd, request_id, msg, docs):
         """A proxy for SocketInfo.write_command that handles event publishing.
         """
         if self.publish:
             duration = datetime.datetime.now() - self.start_time
-            self._start(request_id, docs)
+            self._start(cmd, request_id, docs)
             start = datetime.datetime.now()
         try:
             reply = self.sock_info.write_command(request_id, msg)
             if self.publish:
                 duration = (datetime.datetime.now() - start) + duration
                 self._succeed(request_id, reply, duration)
-        except OperationFailure as exc:
+        except Exception as exc:
             if self.publish:
                 duration = (datetime.datetime.now() - start) + duration
-                self._fail(request_id, exc.details, duration)
+                if isinstance(exc, (NotPrimaryError, OperationFailure)):
+                    failure = exc.details
+                else:
+                    failure = _convert_exception(exc)
+                self._fail(request_id, failure, duration)
             raise
         finally:
             self.start_time = datetime.datetime.now()
         return reply
 
-    def _start(self, request_id, docs):
+    def _start(self, cmd, request_id, docs):
         """Publish a CommandStartedEvent."""
-        cmd = self.command.copy()
         cmd[self.field] = docs
         self.listeners.publish_command_start(
             cmd, self.db_name,
-            request_id, self.sock_info.address, self.op_id)
+            request_id, self.sock_info.address, self.op_id,
+            self.sock_info.service_id)
         return cmd
 
     def _succeed(self, request_id, reply, duration):
         """Publish a CommandSucceededEvent."""
         self.listeners.publish_command_success(
             duration, reply, self.name,
-            request_id, self.sock_info.address, self.op_id)
+            request_id, self.sock_info.address, self.op_id,
+            self.sock_info.service_id)
 
     def _fail(self, request_id, failure, duration):
         """Publish a CommandFailedEvent."""
         self.listeners.publish_command_failure(
             duration, failure, self.name,
-            request_id, self.sock_info.address, self.op_id)
+            request_id, self.sock_info.address, self.op_id,
+            self.sock_info.service_id)
+
+
+# From the Client Side Encryption spec:
+# Because automatic encryption increases the size of commands, the driver
+# MUST split bulk writes at a reduced size limit before undergoing automatic
+# encryption. The write payload MUST be split at 2MiB (2097152).
+_MAX_SPLIT_SIZE_ENC = 2097152
+
+
+class _EncryptedBulkWriteContext(_BulkWriteContext):
+    __slots__ = ()
+
+    def _batch_command(self, cmd, docs):
+        namespace = self.db_name + '.$cmd'
+        msg, to_send = _encode_batched_write_command(
+            namespace, self.op_type, cmd, docs, self.check_keys,
+            self.codec, self)
+        if not to_send:
+            raise InvalidOperation("cannot do an empty bulk write")
+
+        # Chop off the OP_QUERY header to get a properly batched write command.
+        cmd_start = msg.index(b"\x00", 4) + 9
+        cmd = _inflate_bson(memoryview(msg)[cmd_start:],
+                            DEFAULT_RAW_BSON_OPTIONS)
+        return cmd, to_send
+
+    def execute(self, cmd, docs, client):
+        batched_cmd, to_send = self._batch_command(cmd, docs)
+        result = self.sock_info.command(
+            self.db_name, batched_cmd,
+            codec_options=_UNICODE_REPLACE_CODEC_OPTIONS,
+            session=self.session, client=client)
+        return result, to_send
+
+    def execute_unack(self, cmd, docs, client):
+        batched_cmd, to_send = self._batch_command(cmd, docs)
+        self.sock_info.command(
+            self.db_name, batched_cmd, write_concern=WriteConcern(w=0),
+            session=self.session, client=client)
+        return to_send
+
+    @property
+    def max_split_size(self):
+        """Reduce the batch splitting size."""
+        return _MAX_SPLIT_SIZE_ENC
 
 
 def _raise_document_too_large(operation, doc_size, max_size):
@@ -1281,6 +1456,7 @@ def _batched_write_command_impl(
     # Max BSON object size + 16k - 2 bytes for ending NUL bytes.
     # Server guarantees there is enough room: SERVER-10643.
     max_cmd_size = max_bson_size + _COMMAND_OVERHEAD
+    max_split_size = ctx.max_split_size
 
     # No options
     buf.write(_ZERO_32)
@@ -1292,7 +1468,7 @@ def _batched_write_command_impl(
 
     # Where to write command document length
     command_start = buf.tell()
-    buf.write(bson.BSON.encode(command))
+    buf.write(encode(command))
 
     # Start of payload
     buf.seek(-1, 2)
@@ -1313,16 +1489,18 @@ def _batched_write_command_impl(
     for doc in docs:
         # Encode the current operation
         key = b(str(idx))
-        value = bson.BSON.encode(doc, check_keys, opts)
+        value = encode(doc, check_keys, opts)
         # Is there enough room to add this document? max_cmd_size accounts for
         # the two trailing null bytes.
-        enough_data = (buf.tell() + len(key) + len(value)) >= max_cmd_size
+        doc_too_large = len(value) > max_cmd_size
+        if doc_too_large:
+            write_op = list(_FIELD_MAP.keys())[operation]
+            _raise_document_too_large(
+                write_op, len(value), max_bson_size)
+        enough_data = (idx >= 1 and
+                       (buf.tell() + len(key) + len(value)) >= max_split_size)
         enough_documents = (idx >= max_write_batch_size)
         if enough_data or enough_documents:
-            if not idx:
-                write_op = list(_FIELD_MAP.keys())[operation]
-                _raise_document_too_large(
-                    write_op, len(value), max_bson_size)
             break
         buf.write(_BSONOBJ)
         buf.write(key)
@@ -1355,16 +1533,16 @@ class _OpReply(object):
 
     def __init__(self, flags, cursor_id, number_returned, documents):
         self.flags = flags
-        self.cursor_id = cursor_id
+        self.cursor_id = Int64(cursor_id)
         self.number_returned = number_returned
         self.documents = documents
 
-    def raw_response(self, cursor_id=None):
+    def raw_response(self, cursor_id=None, user_fields=None):
         """Check the response header from the database, without decoding BSON.
 
         Check the response for errors and unpack.
 
-        Can raise CursorNotFound, NotMasterError, ExecutionTimeout, or
+        Can raise CursorNotFound, NotPrimaryError, ExecutionTimeout, or
         OperationFailure.
 
         :Parameters:
@@ -1386,8 +1564,8 @@ class _OpReply(object):
             error_object = bson.BSON(self.documents).decode()
             # Fake the ok field if it doesn't exist.
             error_object.setdefault("ok", 0)
-            if error_object["$err"].startswith("not master"):
-                raise NotMasterError(error_object["$err"], error_object)
+            if error_object["$err"].startswith(HelloCompat.LEGACY_ERROR):
+                raise NotPrimaryError(error_object["$err"], error_object)
             elif error_object.get("code") == 50:
                 raise ExecutionTimeout(error_object.get("$err"),
                                        error_object.get("code"),
@@ -1396,16 +1574,19 @@ class _OpReply(object):
                                    error_object.get("$err"),
                                    error_object.get("code"),
                                    error_object)
-        return [self.documents]
+        if self.documents:
+            return [self.documents]
+        return []
 
     def unpack_response(self, cursor_id=None,
-                        codec_options=_UNICODE_REPLACE_CODEC_OPTIONS):
+                        codec_options=_UNICODE_REPLACE_CODEC_OPTIONS,
+                        user_fields=None, legacy_response=False):
         """Unpack a response from the database and decode the BSON document(s).
 
         Check the response for errors and unpack, returning a dictionary
         containing the response data.
 
-        Can raise CursorNotFound, NotMasterError, ExecutionTimeout, or
+        Can raise CursorNotFound, NotPrimaryError, ExecutionTimeout, or
         OperationFailure.
 
         :Parameters:
@@ -1416,7 +1597,10 @@ class _OpReply(object):
             :class:`~bson.codec_options.CodecOptions`
         """
         self.raw_response(cursor_id)
-        return bson.decode_all(self.documents, codec_options)
+        if legacy_response:
+            return bson.decode_all(self.documents, codec_options)
+        return bson._decode_all_selective(
+            self.documents, codec_options, user_fields)
 
     def command_response(self):
         """Unpack a command response."""
@@ -1424,6 +1608,16 @@ class _OpReply(object):
         assert self.number_returned == 1
         return docs[0]
 
+    def raw_command_response(self):
+        """Return the bytes of the command response."""
+        # This should never be called on _OpReply.
+        raise NotImplementedError
+
+    @property
+    def more_to_come(self):
+        """Is the moreToCome bit set on this response?"""
+        return False
+
     @classmethod
     def unpack(cls, msg):
         """Construct an _OpReply from raw bytes."""
@@ -1444,15 +1638,28 @@ class _OpMsg(object):
     UNPACK_FROM = struct.Struct("<IBi").unpack_from
     OP_CODE = 2013
 
+    # Flag bits.
+    CHECKSUM_PRESENT = 1
+    MORE_TO_COME = 1 << 1
+    EXHAUST_ALLOWED = 1 << 16  # Only present on requests.
+
     def __init__(self, flags, payload_document):
         self.flags = flags
         self.payload_document = payload_document
 
-    def raw_response(self, cursor_id=None):
-        raise NotImplementedError
+    def raw_response(self, cursor_id=None, user_fields={}):
+        """
+        cursor_id is ignored
+        user_fields is used to determine which fields must not be decoded
+        """
+        inflated_response = _decode_selective(
+            RawBSONDocument(self.payload_document), user_fields,
+            DEFAULT_RAW_BSON_OPTIONS)
+        return [inflated_response]
 
     def unpack_response(self, cursor_id=None,
-                        codec_options=_UNICODE_REPLACE_CODEC_OPTIONS):
+                        codec_options=_UNICODE_REPLACE_CODEC_OPTIONS,
+                        user_fields=None, legacy_response=False):
         """Unpack a OP_MSG command response.
 
         :Parameters:
@@ -1460,21 +1667,41 @@ class _OpMsg(object):
           - `codec_options` (optional): an instance of
             :class:`~bson.codec_options.CodecOptions`
         """
-        return bson.decode_all(self.payload_document, codec_options)
+        # If _OpMsg is in-use, this cannot be a legacy response.
+        assert not legacy_response
+        return bson._decode_all_selective(
+            self.payload_document, codec_options, user_fields)
 
     def command_response(self):
         """Unpack a command response."""
         return self.unpack_response()[0]
 
+    def raw_command_response(self):
+        """Return the bytes of the command response."""
+        return self.payload_document
+
+    @property
+    def more_to_come(self):
+        """Is the moreToCome bit set on this response?"""
+        return self.flags & self.MORE_TO_COME
+
     @classmethod
     def unpack(cls, msg):
         """Construct an _OpMsg from raw bytes."""
         flags, first_payload_type, first_payload_size = cls.UNPACK_FROM(msg)
         if flags != 0:
-            raise ProtocolError("Unsupported OP_MSG flags (%r)" % (flags,))
+            if flags & cls.CHECKSUM_PRESENT:
+                raise ProtocolError(
+                    "Unsupported OP_MSG flag checksumPresent: "
+                    "0x%x" % (flags,))
+
+            if flags ^ cls.MORE_TO_COME:
+                raise ProtocolError(
+                    "Unsupported OP_MSG flags: 0x%x" % (flags,))
         if first_payload_type != 0:
             raise ProtocolError(
-                "Unsupported OP_MSG payload type (%r)" % (first_payload_type,))
+                "Unsupported OP_MSG payload type: "
+                "0x%x" % (first_payload_type,))
 
         if len(msg) != first_payload_size + 5:
             raise ProtocolError("Unsupported OP_MSG reply: >1 section")
@@ -1492,24 +1719,25 @@ _UNPACK_REPLY = {
 
 
 def _first_batch(sock_info, db, coll, query, ntoreturn,
-                 slave_ok, codec_options, read_preference, cmd, listeners):
+                 secondary_ok, codec_options, read_preference, cmd, listeners):
     """Simple query helper for retrieving a first (and possibly only) batch."""
     query = _Query(
         0, db, coll, 0, query, None, codec_options,
         read_preference, ntoreturn, 0, DEFAULT_READ_CONCERN, None, None,
-        None)
+        None, None, False)
 
     name = next(iter(cmd))
     publish = listeners.enabled_for_commands
     if publish:
         start = datetime.datetime.now()
 
-    request_id, msg, max_doc_size = query.get_message(slave_ok, sock_info)
+    request_id, msg, max_doc_size = query.get_message(secondary_ok, sock_info)
 
     if publish:
         encoding_duration = datetime.datetime.now() - start
         listeners.publish_command_start(
-            cmd, db, request_id, sock_info.address)
+            cmd, db, request_id, sock_info.address,
+            service_id=sock_info.service_id)
         start = datetime.datetime.now()
 
     sock_info.send_message(msg, max_doc_size)
@@ -1519,12 +1747,13 @@ def _first_batch(sock_info, db, coll, query, ntoreturn,
     except Exception as exc:
         if publish:
             duration = (datetime.datetime.now() - start) + encoding_duration
-            if isinstance(exc, (NotMasterError, OperationFailure)):
+            if isinstance(exc, (NotPrimaryError, OperationFailure)):
                 failure = exc.details
             else:
                 failure = _convert_exception(exc)
             listeners.publish_command_failure(
-                duration, failure, name, request_id, sock_info.address)
+                duration, failure, name, request_id, sock_info.address,
+                service_id=sock_info.service_id)
         raise
     # listIndexes
     if 'cursor' in cmd:
@@ -1543,6 +1772,7 @@ def _first_batch(sock_info, db, coll, query, ntoreturn,
     if publish:
         duration = (datetime.datetime.now() - start) + encoding_duration
         listeners.publish_command_success(
-            duration, result, name, request_id, sock_info.address)
+            duration, result, name, request_id, sock_info.address,
+            service_id=sock_info.service_id)
 
     return result
